/*
 *	group menu
 */

#include "config.h"
#include "articles.h"
#include "db.h"
#include "term.h"
#include "menu.h"
#include "keymap.h"
#include "regexp.h"
#ifdef HAVE_SYSLOG
#include <syslog.h>
#endif

export int  dont_split_digests = 0;
export int  dont_sort_articles = 0;
export int  also_cross_postings = 0;
export int  seq_cross_filtering = 0;

import int  article_limit, also_read_articles;
import int  no_update;
import int  merged_menu;
import int  killed_articles;

char *quick_match();

/*
 * completion of group name
 */

group_completion(hbuf, ix)
char *hbuf;
int ix;
{
    static group_number next_group, n1, n2;
    static char *head, *tail, *last;
    static int  tail_offset, prev_lgt, l1, l2;
    static group_header *prev_group, *p1, *p2;
    register group_header *gh;
    register char *t1, *t2;
    int order;

    if (ix < 0) return 0;

    if (hbuf) {
	n2 = next_group = 0;
	p2 = prev_group = NULL;
	l2 = 0;
	
	if (head = strrchr(hbuf, ',')) 
	    head++;
	else
	    head = hbuf;
	tail = hbuf + ix;
	tail_offset = ix - (head - hbuf);
	if (last = strrchr(head, '.')) last++; else last = head;
	return 1;
    }

    if (ix) {
	n1 = next_group, p1 = prev_group, l1 = prev_lgt;
	next_group = n2, prev_group = p2, prev_lgt = l2;
	list_completion((char *)NULL);
    }
    
    *tail = NUL;
    
    while (next_group < master.number_of_groups) {
	gh = sorted_groups[next_group++]; 
	if (gh->group_name_length <= tail_offset) continue;

	if (prev_group &&
	    strncmp(prev_group->group_name, gh->group_name, prev_lgt) == 0)
	    continue;
	
	order = strncmp(gh->group_name, head, tail_offset);
	if (order < 0) continue;
	if (order > 0) break;
	
	t1 = gh->group_name + tail_offset;
	if (t2 = strchr(t1, '.')) {
	    strncpy(tail, t1, t2 - t1 + 1);
	    tail[t2 - t1 + 1] = NUL;
	} else
	    strcpy(tail, t1);

	prev_group = gh;
	prev_lgt = tail_offset + strlen(tail);
	if (ix) {
	    if (list_completion(last) == 0) break;
	} else
	    return 1;
    }

    if (ix) {
	n2 = next_group, p2 = prev_group, l2 = prev_lgt;
	if (n2 > master.number_of_groups) 
	    n2 = 0, p2 = NULL, l2 = 0;
	next_group = n1, prev_group = p1, prev_lgt = l1;
	return 1;
    }

    next_group = 0;
    prev_group = NULL;
    return 0;
}


/* flags to access_group */

#define	ALSO_CROSS_POSTINGS	0x01
#define	DONT_SORT_ARTICLES	0x02
#define	DONT_SPLIT_DIGESTS	0x04	/* only full digest */
#define	ALSO_FULL_DIGEST	0x08	/* also full digest */
#define EXTRA_ARTICLES		0x10	/* add to current menu */

static access_group(gh, first_article, last_article, flags, submask, do_kill)
register group_header	*gh;
article_number	first_article, last_article;
int		flags;
char 		*submask;
int		do_kill;
{
    group_header		*cpgh;
    FILE			*data;
    data_header			hdr;
    off_t	  		data_offset;
    register article_header	*ah;
    cross_post_number		cross_post;
    int				skip_digest;
    int				n;
    string_marker		str_marker;
    memory_marker		mem_marker;
    article_number		art_num;
    static regexp		*subpattern = NULL;
    static char			subptext[80];
    
    if (first_article < gh->first_l_article) 
	first_article = gh->first_l_article;

    if (last_article > gh->last_l_article) 
	last_article = gh->last_l_article;

    if (last_article == 0 || first_article > last_article) return 0;

    data = open_data_file(gh, 'd', OPEN_READ);
    if (data == NULL) return -1;

    if ((data_offset = get_data_offset(gh, first_article)) == (off_t)(-1))
	return -1;

    if (submask && *submask == '/') {
	submask++;
	if (subpattern != NULL) {
	    if (strncmp(submask, subptext, 80) != 0) {
		free((char *)subpattern);
		subpattern = NULL;
	    }
	}
	if (subpattern == NULL) {
	    strncpy(subptext, submask, 80);
	    subpattern = regcomp(submask);
	    if (subpattern == NULL) return -1;
	}
	submask = NULL;
    } else
	if (subpattern != NULL) {
	    free((char *)subpattern);
	    subpattern = NULL;
	}
    
    if ((flags & EXTRA_ARTICLES) == 0)
	mark_memory(&mem_marker);
    
    ah = alloc_art();

    skip_digest = 0;

skip_to_next:
    fseek(data, data_offset, 0);
    
read_next:
    if (data_offset >= gh->data_write_offset) goto out;
    
    if (!db_read_art(data, &hdr, &data_offset)) goto out;

    if (hdr.dh_lpos == (off_t)0)	/* article not accessible */
	goto read_next;

    if (art_num = hdr.dh_number) {
	if (art_num < 0) art_num = -art_num;
	if (art_num > gh->last_l_article ||
	    art_num < gh->first_l_article)
	    goto data_error;
    }
    
    data_offset += hdr.dh_cross_postings * sizeof(cross_post_number)
		 + hdr.dh_sender_length 
		 + hdr.dh_subject_length;

    if (skip_digest && IS_SUB_DIGEST(hdr)) goto skip_to_next;
	
    skip_digest = 0;
	
    if (hdr.dh_cross_postings) {
	if ((flags & ALSO_CROSS_POSTINGS) == 0) {
	    n = hdr.dh_cross_postings;
	    do {
		if (fread((char *)&cross_post, sizeof(cross_post_number), 1, data) != 1)
		    goto data_error;
#ifdef NETWORK_DATABASE
#ifndef NETWORK_BYTE_ORDER
		cross_post = ntohl(cross_post);
#endif
#endif		
		cpgh = &active_groups[cross_post];
		if (cpgh == gh) {
		    if (seq_cross_filtering) continue;
		    if (n > 1)
			fseek(data, (off_t)(sizeof(cross_post_number)*(n-1)), 1);
		    n = 0;
		    break;
		}
		if (cpgh->group_flag & G_SUBSCRIPTION) {
		    if (!seq_cross_filtering) break;
		    if (cpgh->preseq_index > 0 &&
			cpgh->preseq_index < gh->preseq_index) break;
		}
	    } while (--n > 0);
	    if (n > 0) {
		if (IS_DIGEST_HEADER(hdr)) skip_digest++;
		goto skip_to_next;
	    } 
	} else
	    fseek(data, (off_t)(sizeof(cross_post_number)*hdr.dh_cross_postings), 1);
    }	

    ah->flag = 0;

    if (IS_DIGEST_HEADER(hdr)) {
	if ((flags & (DONT_SPLIT_DIGESTS | ALSO_FULL_DIGEST)) == 0) 
	    goto skip_to_next;	/* don't want the full digest when split */
	skip_digest++;
	ah->flag |= A_FULL_DIGEST;
    } else
	if (IS_SUB_DIGEST(hdr))
	    ah->flag |= A_DIGEST;
    
    ah->a_number = ARTICLE_NUMBER(hdr);
    if (ah->a_number > last_article) goto out;
    
    mark_str(&str_marker);
    
    if (ah->name_length = hdr.dh_sender_length) {
	ah->sender = alloc_str((int)hdr.dh_sender_length);
	if (fread(ah->sender, sizeof(char), (int)hdr.dh_sender_length, data)
	    != hdr.dh_sender_length) goto data_error;
    } else
	ah->sender = "";
    
    if (ah->subj_length = hdr.dh_subject_length) {
	ah->subject = alloc_str((int)hdr.dh_subject_length);
	if (fread(ah->subject, sizeof(char), (int)hdr.dh_subject_length, data)
	    !=  hdr.dh_subject_length) goto data_error;
    } else
	ah->subject = "";

    ah->hpos	= hdr.dh_hpos;
    ah->fpos	= ah->hpos + (off_t)(hdr.dh_fpos);
    ah->lpos	= hdr.dh_lpos;
    
    ah->replies	= hdr.dh_replies;
    ah->lines	= hdr.dh_lines;

    ah->t_stamp = hdr.dh_date;
    
    if ((subpattern && !regexec(subpattern, ah->subject)) ||
	(submask && quick_match(ah->subject, submask) == NULL) ||
	(do_kill && kill_article(ah))) {
	release_str(&str_marker);
	goto read_next;
    }

    ah->a_group = merged_menu ? current_group : NULL;
    
    add_article(ah);
    ah = alloc_art();

    goto read_next;

    
data_error:
    log_entry('E', "%s: data inconsistency", gh->group_name);
    fclose(data);
    if ((flags & EXTRA_ARTICLES) == 0)
	release_memory(&mem_marker);
    return -1;
    
out:
    fclose(data);
    
    if ((flags & DONT_SORT_ARTICLES) == 0)
	sort_articles();
    
    return n_articles > 0 ? 1 : 0;
}

static article_number current_first_article;


static print_header()
{
    extern long unread_articles;
    extern int unread_groups;
    
    so_printxy(0, 0, "Newsgroup: %s", current_group->group_name);
    clrline();
    
    so_gotoxy(-1, 0, 0);
    
    so_printf("Articles: %d", n_articles);

    if (no_update) {
	so_printf(" NO UPDATE");
    } else {
	if (current_first_article > current_group->first_article + 1)
	    so_printf((killed_articles > 0) ? "(L-%ld,K-%d)" : "(L-%ld)", 
		      current_first_article - current_group->first_article - 1,
		      killed_articles);
	else
	    if (killed_articles > 0)
		so_printf(" (K-%d)", killed_articles);
	
	if (unread_articles > 0) {
	    so_printf(" of %ld", unread_articles);
	    if (unread_groups > 0) 
		so_printf("/%d", unread_groups);
	}
	
	if ((current_group->group_flag & G_SUBSCRIPTION) == 0)
	    so_printf(" UNSUB");
	else if (current_group->group_flag & G_NEW) 
	    so_printf(" NEW"); 
	
	if (current_group->group_flag & G_READ) 
	    so_printf(" READ");
    }
    
    so_end();

    return 1;	/* number of lines in header */
}



group_menu(gh, first_art, submask, do_kill, menu)
register group_header *gh;
article_number first_art;
char *submask;
int do_kill;
int (*menu)();
{
    int status, was_unread;
    int menu_cmd, access_mode, did_selection, o_killed;
    article_number o_first_article, prev_last, last_article;
    memory_marker sel_marker;	/* for marking old selection */
    
#define	menu_return(cmd) { menu_cmd = (cmd); goto menu_exit; }
    
    o_first_article = current_first_article;
    o_killed = killed_articles;
    mark_memory(&sel_marker);
    mark_var_stack();
    
    if (init_group(gh) <= 0) menu_return( ME_NEXT );
    m_invoke(-1);
    
after_selection:

    did_selection = 0;
    killed_articles = 0;

    /* don't lose (a few) new articles when reentering a read group */
    /* (the user will expect the menu to be the same, and may overlook */
    /* the new articles) */
    if (gh->group_flag & G_READ) goto update_unsafe;
    
    prev_last = gh->last_l_article;

    was_unread = add_unread(gh, -1);
    
    if (update_group(gh) < 0) {
	status = -2;
	goto access_exception;
    }
    
    if (gh->last_l_article < prev_last) {
	/* expire + renumbering */
	int32 updflag;

	if ((gh->last_article = gh->first_l_article - 1) < 0)
	    gh->last_article = 0;
	gh->first_article = gh->last_article;
	updflag = gh->group_flag & G_READ;
	gh->group_flag &= ~G_READ;
	update_rc(gh);
	gh->group_flag &= ~G_READ;
	gh->group_flag |= updflag;
    }

    if (was_unread)
	add_unread(gh, 1);

 update_unsafe:
    gotoxy(0, 0);
    fl;

    access_mode = 0;
    if (also_read_articles || submask || also_cross_postings)
	access_mode |= ALSO_CROSS_POSTINGS;
    if (dont_split_digests)
	access_mode |= DONT_SPLIT_DIGESTS;
    if (dont_sort_articles)
	access_mode |= DONT_SORT_ARTICLES;
    
    if (first_art == -1) {
	if (submask == NULL && !also_read_articles) {
	    if (has_selection(gh, &current_first_article, &last_article)) {
		status = access_group(gh, current_first_article, last_article,
				      access_mode | DONT_SORT_ARTICLES, (char *)NULL, 0);
		do_selections(status >= 0 && n_articles);
		if (status < 0) goto access_exception;
		if (n_articles) {
		    did_selection = 1;
		    if (!dont_sort_articles) sort_articles();
		    goto read_the_articles;
		}
	    }
	}
	
	if (also_read_articles || submask)
	    current_first_article = gh->first_l_article;
	else
	    current_first_article = gh->last_article + 1;
	
	if (article_limit > 0) {
	    article_number temp;
	    
	    temp = gh->last_l_article - article_limit + 1;
	    if (current_first_article < temp) current_first_article = temp;
	}
    } else
	current_first_article = first_art;
    
    status = access_group(gh, current_first_article, gh->last_l_article,
			  access_mode, submask, do_kill);

 access_exception:

    if (status < 0) {
	if (status == -1) {
	    clrdisp();
	    msg("DATABASE CORRUPTED FOR GROUP %s", gh->group_name);
#ifdef HAVE_SYSLOG
	    openlog("nn", LOG_CONS, LOG_DAEMON);
	    syslog(LOG_ALERT, "database corrupted for newsgroup %s.",
		   gh->group_name);
	    closelog();
#endif
	    user_delay(5);
	}
	menu_return( ME_NEXT );
    }
    
    if (n_articles == 0) {
	m_break_entry();
	menu_return( ME_NO_ARTICLES );
    }
    
 read_the_articles:

    menu_cmd = (*menu)(print_header);
    
 menu_exit:

    if (menu_cmd == ME_QUIT || menu_cmd == ME_NEXT || menu_cmd == ME_PREV)
	if (submask == NULL && !no_update) 
	    save_selection(gh, current_first_article, gh->last_l_article, 0);

    if (menu_cmd == ME_READ || menu_cmd == ME_NO_ARTICLES) {
	if (did_selection) {
	    int32 was_read = gh->group_flag & G_READ;

	    prev_last = gh->last_l_article;
	    gh->last_l_article = last_article;
	    update_rc(gh);
	    gh->last_l_article = prev_last;
	    
	    if (last_article < gh->last_l_article) {
		gh->group_flag &= ~G_READ;
		gh->group_flag |= was_read;
		release_memory(&sel_marker);
		goto after_selection;
	    }
	} else
	    if (submask == NULL && !also_read_articles &&
		(first_art == -1 || current_first_article == gh->first_article + 1)) {
		if (menu_cmd != ME_NO_ARTICLES)
		    gh->group_flag &= ~G_NEW;
		update_rc(gh);
	    }
    }
    
    current_first_article = o_first_article;
    killed_articles = o_killed;
    restore_variables();
    
    return menu_cmd;
}


goto_group(command, ah)
int command;
article_header *ah;
{
    register group_header *gh;
    char ans1, *answer, *submask, buffer[FILENAME];
    article_number first, o_current_first;
    memory_marker mem_marker;
    group_header *orig_group;
    int menu_cmd;
    extern int menu(), file_completion();
    extern article_header *get_menu_article();
    extern int get_from_macro;
    extern group_header *jump_to_group;

#define goto_return( cmd ) \
    { menu_cmd = cmd; goto goto_exit; }

    m_startinput();
    
    gh = orig_group = current_group;
    
    o_current_first = current_first_article;
    
    submask = NULL;

    if (command == K_GOTO_GROUP) 
	goto get_group_name;
    
    for (;;) {
	if (command == K_ADVANCE_GROUP)
	    gh = gh->next_group;
	else
	    gh = gh->prev_group;
	
	if (gh == NULL)
	    goto_return(ME_NO_REDRAW);
	
	if (gh->first_l_article >= gh->last_l_article) continue;
	
	prompt("\1Enter\1 %s%s%s ?", gh->group_name,
	       gh->group_flag & G_SUBSCRIPTION ? "" : " (UNSUB)",
	       gh->group_flag & G_READ ? " (READ)" : "" );
	
	command = get_c();
	if (command & GETC_COMMAND) goto_return(ME_REDRAW);
	if (command == 'y' || command == 'Y') break;
	if (command == 'n' || command == 'N') goto_return(ME_NO_REDRAW);
	
	command = menu_key_map[command];
	if (command == K_CONTINUE) break;
	if (command == K_ADVANCE_GROUP) continue;
	if (command == K_BACK_GROUP) continue;
	if (command == K_GOTO_GROUP) goto get_group_name;
	goto_return(ME_NO_REDRAW);
    }
    
    if (gh == orig_group) goto_return(ME_NO_REDRAW);
    
    goto get_first;
    
 get_group_name:

    if (current_group == NULL) {
	prompt("\1Enter Group or Folder\1 (+./~) ");
	answer = get_s(NONE, NONE, "+./~", group_completion);
    } else {
	prompt("\1Group or Folder\1 (+./~ %=N) ");
	answer = get_s(NONE, NONE, "+./0123456789~=%", group_completion);
    }
    
    if (answer == NULL) goto_return(ME_NO_REDRAW);

    if ((ans1 = *answer) == NUL) {
	if (current_group == NULL) goto_return(ME_NO_REDRAW);
	goto get_first;
    }
    
    sprintf(buffer, "%c", ans1);
    
    switch (ans1) {

     case '%':
	if (current_group == NULL) goto_return(ME_NO_REDRAW);
	if ((current_group->group_flag & G_FOLDER) == 0) {
	    if (!ah) {
		prompt("\1READ\1");
		if ((ah = get_menu_article()) == NULL) 
		    goto_return(ME_NO_REDRAW);
	    }
	    if ((ah->flag & A_DIGEST) == 0) {
		*group_file_name = NUL;
		sprintf(buffer, "%s%ld", group_path_name, ah->a_number);
		answer = buffer;
		goto get_folder;
	    }
	}
	
	msg("cannot split articles inside a folder or digest");
	goto_return(ME_NO_REDRAW);

     case '.':
     case '~':
	strcat(buffer, "/");
     case '+':
     case '/':
	if (!get_from_macro) {
	    prompt("\1Folder\1 ");
	    answer = get_s(NONE, buffer, NONE, file_completion);
	    if (answer == NULL || answer[0] == NUL) goto_return(ME_NO_REDRAW);
	}
	
     get_folder:
	m_endinput();
	if (!expand_file_name(buffer, answer)) goto_return (ME_NO_REDRAW);
	menu_cmd = folder_menu(buffer);
	init_group(orig_group);
	goto goto_exit;

	
     case 'a':
	if (answer[1] == NUL || strcmp(answer, "all") == 0) {
	    if (current_group == NULL) goto_return(ME_NO_REDRAW);
	    first = 0;
	    goto more_articles;
	}
	break;
	
     case 'u':
	if (answer[1] == NUL || strcmp(answer, "unread") == 0) {
	    if (current_group == NULL) goto_return(ME_NO_REDRAW);
	    first = gh->first_article + 1;
	    goto more_articles;
	}
	break;

     case 's':
	if (answer[1] == NUL) {
	    if (current_group == NULL) goto_return(ME_NO_REDRAW);
	    goto get_subject;
	}
	
	break;
	
     case '=':
	if (current_group == NULL) goto_return(ME_NO_REDRAW);
	goto get_subject;
	
     case '0':
     case '1':
     case '2':
     case '3':
     case '4':
     case '5':
     case '6':
     case '7':
     case '8':
     case '9':
	if (current_group == NULL) goto_return(ME_NO_REDRAW);
	if (current_first_article <= gh->first_l_article) {
	    msg("No extra articles");
	    flush_input();
	    goto_return(ME_NO_REDRAW);
	}
	
	if (!get_from_macro) {
	    prompt("\1Number of extra articles\1 max %ld: ",
		   current_first_article - gh->first_l_article);
	    sprintf(buffer, "%c", ans1);
	    answer = get_s(NONE, buffer, NONE, NO_COMPLETION);
	    if (answer == NULL || *answer ==  NUL) goto_return(ME_NO_REDRAW);
	}
	
	first = current_first_article - atol(answer);
	goto more_articles;
	
     default:
	break;
    }

    if ((gh = lookup(answer)) == NULL) {
	msg("No group named %s", answer);
	goto_return(ME_NO_REDRAW);
    }


 get_first:

    m_advinput();

    if (gh->group_flag & G_BLOCKED ||
	gh->last_l_article == 0 ||
	gh->last_l_article < gh->first_l_article) {
	msg("Group %s is %s", answer, 
	    (gh->group_flag & G_BLOCKED) ? "blocked" : "empty");
	goto_return(ME_NO_REDRAW);
    }

    if (gh == orig_group
	&& current_first_article <= gh->first_article) {

	if (current_first_article <= gh->first_l_article) {
	    /* no more articles to read */
	    *answer = '=';
	    get_from_macro = 0;
	    goto get_subject;
	}
	    
	prompt("\1Number of extra articles\1 all %ld, =subject: ",
	       (long)(current_first_article - gh->first_l_article));

	answer = "a=s";
    } else {
	if (gh != orig_group)
	    current_first_article = gh->last_l_article + 1;
	
        prompt("\1Number of%s articles\1 u)nread %ld,%s a)ll %ld, s)ubject: ",
	       gh == orig_group ? " extra" : "",
	       (long)(current_first_article - gh->first_article - 1),
	       (gh->group_flag & G_UNREAD_COUNT) ? " j)ump," : "",
	       (long)(current_first_article - gh->first_l_article));

	answer = (gh->group_flag & G_UNREAD_COUNT) ? "uja=s" : "ua=s";
    }
    
    answer = get_s(NONE, NONE, answer, NO_COMPLETION);
    if (answer == NULL) goto_return(ME_NO_REDRAW);

get_subject:	/* when *answer == '=' */
    
    switch (*answer) {
	
     case NUL:
	if (gh != orig_group || current_first_article <= gh->first_l_article) {
	    first = 0;
	    break;
	}
	/* else fall thru */
	
     case 'u':
	first = gh->first_article + 1;
	break;

     case 'j':
	jump_to_group = gh;
	goto_return(ME_QUIT);
	
     case 'a':
	first = 0;
	break;
	
     case 's':
     case '=':
	first = 0;
	if (get_from_macro) {
	    submask = answer + 1;
	} else {
	    prompt("=");
	    submask = get_s(ah ? ah->subject : NONE, NONE, ah ? NONE : "%=",
			    NO_COMPLETION);
	    if (submask == NULL) goto_return(ME_NO_REDRAW);
	    if (*submask == '%' || *submask == '=') {
		if ((ah = get_menu_article()) == 0) goto_return(ME_NO_REDRAW);
		*submask = NUL;
	    }
	}
	
	if (*submask == NUL)
	    if (ah) {
		strncpy(submask, ah->subject, GET_S_BUFFER);
		submask[GET_S_BUFFER-1] = NUL;
	    } else
		goto_return(ME_NO_REDRAW);
	
	if (*submask) {
	    if (*submask != '/') init_quick_match(submask);
	} else
	    submask = NULL;
	break;
	
     case '0':
     case '1':
     case '2':
     case '3':
     case '4':
     case '5':
     case '6':
     case '7':
     case '8':
     case '9':
	first = current_first_article - atol(answer);
	break;

     default:
	ding();
	goto_return(ME_NO_REDRAW);
    }
    
    if (first > gh->last_l_article) goto_return(ME_NO_REDRAW);

more_articles:
    if (gh == orig_group && submask == NULL && 
	first < current_first_article) {
	if (current_first_article <= gh->first_l_article) {
	    msg("No extra articles");
	    goto_return(ME_NO_REDRAW);
	}
	
	if (access_group(gh, first, current_first_article - 1,
			 EXTRA_ARTICLES|ALSO_CROSS_POSTINGS, (char *)NULL, 0) < 0) {
	    msg("Cannot read extra articles (now)");
	    goto_return(ME_NO_REDRAW);
	}

	current_first_article = first;
	
	goto_return(ME_REDRAW);
    }

    mark_memory(&mem_marker);
    m_endinput();
    menu_cmd = group_menu(gh, first, submask, 0, menu);
    release_memory(&mem_marker);

    if (gh != orig_group) {
	current_first_article = o_current_first;
	if (orig_group) init_group(orig_group);
    }
    
goto_exit:
    m_endinput();
    return menu_cmd;
}

static merged_header()
{
    so_printxy(0, 0, "MERGED NEWS GROUPS:  %d ARTICLES", n_articles);
    clrline();

    return 1;
}

merge_and_read(submask, do_kill)
char *submask;
int do_kill;
{
    register group_header *cur;
    group_header dummy_group;
    int access_mode;
    
    free_memory();

    access_mode = DONT_SORT_ARTICLES;
    if (also_read_articles || submask || also_cross_postings)
	access_mode |= ALSO_CROSS_POSTINGS;
    if (dont_split_digests)
	access_mode |= DONT_SPLIT_DIGESTS;
	
    Loop_Groups_Sequence(cur) {
	if (s_hangup || s_keyboard) break;

	if (cur->group_flag & G_FOLDER) {
	    printf("\n\rIgnoring folder: %s\n\r", cur->group_name);
	    continue;
	}

	if (!also_read_articles)
	    if (cur->last_article >= cur->last_l_article)
		continue;
	
	printf("\r%s", cur->group_name); clrline();
	
	if (init_group(cur) <= 0) continue;
	access_group(cur, (article_number)(-1), cur->last_l_article, access_mode, submask, do_kill);
    }
    merge_memory();
    if (n_articles == 0) return;
    if (!dont_sort_articles) sort_articles();

    dummy_group.group_flag = 0;
    dummy_group.save_file = NULL;
    dummy_group.group_name = "dummy";
    dummy_group.kill_list = NULL;

    current_group = &dummy_group;
    
    menu(merged_header);
    
    free_memory();
}

unsubscribe(gh)
group_header *gh;
{
    if (no_update) {
	msg("nn started in \"no update\" mode");
	return 0;
    }
    
    if (gh->group_flag & G_FOLDER) {
	msg("cannot unsubscribe to a folder");
	return 0;
    }

    if (gh->group_flag & G_SUBSCRIPTION) {
	prompt("\1Unsubscribe to\1 %s ? ", gh->group_name);
	if (yes(0) <= 0) return 0;

	add_unread(gh, -1);
	gh->group_flag &= ~(G_SUBSCRIPTION | G_NEW);
	write_rc_entry(gh, 0);
	return 1;
    }

    prompt("Already unsubscribed.  \1Resubscribe to\1 %s ? ",
	   gh->group_name);
    if (yes(1) <= 0) return 0;
	
    gh->group_flag |= G_SUBSCRIPTION;
    write_rc_entry(gh, 0);

    add_unread(gh, 1);

    return 1;
}


group_overview(amount)
int amount;	/* 0=>unread,subscribed, 1=>unread,all, 2=>all,all 3=>unsub*/
{
    register group_header *gh;
    
    clrdisp();

    pg_init(0, 2);

    if (amount < 0) {
	Loop_Groups_Sequence(gh) {
	    if (gh->group_flag & (G_READ | G_NO_DIRECTORY))
		continue;
	    if ((gh->group_flag & G_SUBSCRIPTION) == 0)
		continue;
	    if (gh->last_article >= gh->last_l_article)
		continue;
	    if (disp_group(gh) < 0) break;
	}
    } else
    Loop_Groups_Sorted(gh) {

	if (gh->group_flag & G_NO_DIRECTORY) continue;

	if (amount <= 1 && gh->last_article >= gh->last_l_article)
		continue;

	if (amount == 0 && (gh->group_flag & G_SUBSCRIPTION) == 0)
		continue;

	if (amount == 3 && (gh->group_flag & G_SUBSCRIPTION))
		continue;

	if (disp_group(gh) < 0) break;
    }

    pg_end();
}

static disp_group(gh)
group_header *gh;
{
    if (pg_next() < 0) return -1;
    
    printf("%6ld%c%s%s",
	   (long)(gh->last_l_article - gh->last_article),
	   (gh == current_group) ? '*' : ' ',
	   gh->group_name,
	   (gh->group_flag & G_SUBSCRIPTION) == 0 ? " (!)" :
	   gh->enter_macro ? " %" : "");

    return 0;
}
    
